/* 
 *           DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *                    Version 2, December 2004
 *  
 * Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>
 * 
 * Everyone is permitted to copy and distribute verbatim or modified
 * copies of this license document, and changing it is allowed as long
 * as the name is changed.
 *  
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 *  
 *  0. You just DO WHAT THE FUCK YOU WANT TO.
 */

/*
 * Marchen Maze data encoder (used for tiles and sprites)
 * author : MooZ
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#define NIL 0
#define RAW 1
#define RLE 2
#define XOR 3

#define MIN_ZERO_COUNT 5

/* Count bytes equal to zero in a 32 bytes array */
void countZeroes(uint8_t *data, int *count)
{
	int i;
	*count = 0;
	for(i=0; i<32; ++i, ++data)
	{
		if(!*data) ++(*count);
	}
}

/* Choose the encoding type that will maximise the number of '0'.
 * The heuristic is quite simple. Count the number of '0' in the
 * "raw" and "xor" data. If the whole buffer is empty use special
 * NIL encoding. If it's less that a certain threshold, use RLE
 * encoding. Otherwise keep the data uncompressed. */
int computeEncodingType( uint8_t *data, uint8_t *xor )
{
	int i;
	int count[2];
	/* Pick the "encoding" with the most number of '0' */
	/* 1. just count in the raw data */
	countZeroes(data, count);

	/* 2. the raw data buffer is empty */
	if( count[0] == 32 )
		return NIL;
	
	/* 3. perform interleaved xor and count */
	memcpy(xor, data, 32 * sizeof(uint8_t));
	for(i=0; i<14; i+=2)
	{               
		xor[i+ 2] ^= data[i   ];
		xor[i+ 3] ^= data[i+ 1];
		xor[i+18] ^= data[i+16];
		xor[i+19] ^= data[i+17];
	}
	countZeroes(xor, count+1);

	/* 4. choose the encoding scheme that maximize the number of '0' */
	if(count[1] >= count[0])
		if(count[1] >= MIN_ZERO_COUNT)
			return XOR;
	if( count[0] >= MIN_ZERO_COUNT)
		return RLE;
	return RAW;
}

/* */
void rleEncode( uint8_t *src, uint8_t *dest, int* len )
{
	int i, j, k;
	/* 0 byte compactation. The header is a bitmask where a bit equals
 	 * to 1 means that the byte and non null and must be fetched from
	 * the compressed data array. */
	for(i=0, k=0, *len=0; i<4; ++i)
	{
			dest[i] = 0;
			for(j=0; j<8; ++j, ++k)
			{
					if(src[k])
					{
							dest[i] |= (1 << j);
							dest[*len+4]  = src[k];
							++(*len);
					}
			}
	}
	*len += 4;
}

/* Let's go! */
int main( int argc, char **argv )
{
	uint8_t src[32], xor[32], rle[32];
	uint8_t *dest, header;
	int srcSize, itemCount, destSize, encoding;
	int i, j, err = 1;

	FILE *input, *output, *headerOutput;
	char *headerFilename;

	/* Check command line arguments */
	if( argc != 3 )
	{
		fprintf( stderr, "Usage: %s input output\n", argv[0] );
		return 1;
	}

	/* Generate filename for headers */
	headerFilename = (char*)malloc( sizeof(char) * strlen(argv[2])+5);
	if( headerFilename == NULL )
	{
		perror(strerror(errno));
		return 1;
	}
	sprintf( headerFilename, "%s.hed", argv[2] );

	/* Open input file */
	input = fopen( argv[1], "rb" );
	if( input == NULL )
	{
		fprintf( stderr, "Unable to open %s: %s\n", argv[1], strerror(errno));
		goto err_0;
	}

	/* Get input size */
	fseek( input, 0, SEEK_END );
	srcSize = ftell( input );
	fseek( input, 0, SEEK_SET );
	srcSize -= ftell( input );

	/* Check size */
	itemCount = srcSize / 32;
	if( itemCount <= 0 )
	{
		fprintf( stderr, "%s is too small (%d bytes)\n", argv[1], srcSize );
		goto err_0;
	}

	/* It's better if the size is a multiple of 32*4 (128) */
	if( srcSize & 127 )
	{
		fprintf( stderr, "Warning! %s size is not a multiple of 128 (4*32 bytes)!\n", argv[1] );
	}

	/* Open output file */
	output = fopen( argv[2], "wb" );
	if( output == NULL )
	{
		fprintf( stderr, "Unable to open %s: %s\n", argv[2], strerror(errno));
		goto err_1;
	}
	
	/* Open header output file */
	headerOutput = fopen( headerFilename, "wb");
	if( headerOutput == NULL )
	{
		fprintf( stderr, "Unable to open %s: %s\n", headerFilename, strerror(errno));
		goto err_2;
	}

	header = 0;
	j = 0;
	for(i=0; i<itemCount; ++i)
	{
		fread( src, 1, 32, input );
		encoding = computeEncodingType(src, xor);
		/* TODO : explain header structure */	
		header |= encoding << (j*2);
		++j;
		if( j == 4 )
		{
			fwrite( &header, 1, 1, headerOutput );
			header = 0;
			j = 0;
		}

		if( encoding != NIL )
		{
			switch( encoding )
			{
				case RAW:
					dest = src;
					destSize = 32;
					break;
				case XOR:
					dest = rle;
					rleEncode(xor, rle, &destSize);
					break;
				case RLE:
					dest = rle;
					rleEncode(src, rle, &destSize);
					break;
			}
			fwrite( dest, 1, destSize, output );
		 }
	}

	/* Flush header */
	if( j )
	{
		fwrite( &header, 1, 1, headerOutput );
	}
	
	err = 1;
err_2:
	fclose( headerOutput );
err_1:
	fclose( output );
err_0:
	fclose( input );
	free( headerFilename );
	return err;
}
